/*
sim-pval.cc

Use a simulation to estimate the probability of an as-or-more-extreme outcome
of a vote, given the vote's results and the mean, stddev of the voting curve.
Creates a matlab file with the data, which can plot the results.

(requires boost library)
Compile:  g++ sim-pval.cc -O2 -o sim-pval
Run:      ./sim-pval
Results:  in matlab, run the generated file simpval.m


Bo Waggoner

Created: 2011-07-26
Modified: 2012-04-15

This information is free.
*/


/*
Parameters:
-output file name: OUTFILE_NAME
-election outcome: v[] = {nA,nB}
-maximum votes castable by one person: maxv
-individual vote distribution: c[]
-number of alpha values: numalpha
-alpha values to use: alphalist[]
-maximum number of supporters to simulate: maxn
-number of trials at each n: trials
-error bound: BE_THRESHOLD
     (When simulating, if error bound is less than BE_THRESHOLD, use Gaussian
     approximation; if greater, simulate each voter. Set to zero to always use
     simulation; set to one to always use approximation.)
-random generator seed: seed

Procedure:
for each alpha in alphalist:
    compute adjust margin of victory beta
    for number of supporters n from 2 to maxn by 2:
        compute error bound; if small, use approximation for next step
        compute Pr[adjusted margin of victory >= beta | nA = nB = n/2]
        
print matlab file OUTFILE_NAME with results

*/


#include <vector>
#include <fstream>
#include <cmath>
#include <ctime>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>


// output filename
const char* OUTFILE_NAME = "simpval.m";

// number of alphas to use and their values
const int numalpha = 3;
const double alphalist[] = {0.2, 0.5, 0.8};

// maximum number of supporters to simulate
const int maxn = 200;

// number of votes observed for A, B
const double v[] = {41,35};

// maximum number of votes to cast
const int maxv = 3;

// vote distribution c[i] = Pr[individual casts i votes]
const double c[maxv+1] = {0.1, 0.5, 0.2,0.2};

// properties of c[] to be computed at beginning of program
double c_cdf[maxv+1];
double mu;
double sigma;
double rho;

// number of trials to run at each n;
// probability = (number that exceed beta)/trials
const int trials = 1000;//1000000;

// random generator seed (set to -1 to use time(0))
int seed = 100;

// a universal constant used in berry-esseen formula
const double BE_C = .4784;

// error threshold
const double BE_THRESHOLD = 0; //.05;


// compute the cdf, mean, etc of c[]
void calcCurveProperties()
{
    double sum = 0;
    for(int i=0; i<=maxv; i++) sum += c[i] * i;
    mu = sum;
    sum = 0;
    for(int i=0; i<=maxv; i++) sum += (i - mu)*(i - mu) * c[i];
    sigma = sqrt(sum);
    sum = 0;
    for(int i=0; i<=maxv; i++)
    {
        double diff = (i-mu);
        diff *= diff*diff * c[i];
        sum += (diff > 0 ? diff : -diff);
    }
    rho = sum;
    c_cdf[0] = c[0];
    for(int i=1; i<=maxv; i++)
    {
        c_cdf[i] = c_cdf[i-1] + c[i];
    }

    //std::cout << mu << "  " << sigma << "  " << rho << "\n";
}

// inverse CDF method
// given a uniform [0,1] random variable, return the number of votes cast
// For large c[], this would be better implemented as a binary search,
// but for arrays of small size, the overhead isn't worth it
int getVotesCast(double cdfgenerated)
{
    int votescast = 0;
    while(votescast < maxv && cdfgenerated > c_cdf[votescast]) votescast++;
    return votescast;
}


// get the probability of exceeding beta when there are n voters
// Approx: use gaussian approximation centered at mu*n with variance sigma^2 * n
//   for number of votes received by each alternative
double getApproxProb(boost::mt19937 &randgen, int n, double alpha, double beta)
{
    // votes for candidate j are approximately normal around mu*n/2, sigma*sqrt(n/2)
    boost::normal_distribution<> normA((double)n*mu/2.0f,sigma*sqrt((double)n/2.0f));
    boost::normal_distribution<> normB((double)n*mu/2.0f,sigma*sqrt((double)n/2.0f));
    
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<> > genA(randgen, normA);
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<> > genB(randgen, normB);
    
    double sum = 0;
    for(int t=0; t<trials; t++)
    {
        double vA = genA();
        double vB = genB();
        
        double b = (vA - vB)/(pow(vA + vB,alpha));
        if(b >= beta) sum++;
    }
    return sum/(double)trials;
}

// get the probability of exceeding beta when there are n voters
// Exact: simulate every voter 
double getExactProb(boost::mt19937 &randgen, int n, double alpha, double beta)
{
    boost::uniform_real<> votercdf(0.0f,1.0f);
    boost::variate_generator<boost::mt19937 &, boost::uniform_real<> > votergencdf(randgen, votercdf);
    
    double sum = 0;
    for(int t=0; t<trials; t++)
    {
        int vA = 0;
        for(int pA = 0; pA < n/2; pA++)
        {
            vA += getVotesCast(votergencdf());
        }
        int vB = 0;
        for(int pB = 0; pB < n/2; pB++)
        {
            vB += getVotesCast(votergencdf());
        }
        double b = ((double)(vA - vB))/(pow(vA + vB,alpha));
        if(b >= beta) sum++;
    }

    return sum/(double)trials;
}

int main()
{
    calcCurveProperties();
    
    // pwrong[a][n] is the (simulated) probability that, using alpha=a, n voters
    //   would have gotten the wrong result or worse
    double* pwrongs[numalpha];
    for(int i=0; i<numalpha; i++) {pwrongs[i] = new double[maxn/2]; pwrongs[i][0] = 0.5;}
    
    // get a mersenne twister generator to use with the normal distributions
    if(seed < 0) seed = time(0);
    boost::mt19937 randgen(seed);
    
    double beta[numalpha];
    for(int i=0; i<numalpha; i++)
    {
        // compute the actual outcome of our election
        beta[i] = (v[0] - v[1])/(pow(v[0] + v[1],alphalist[i]));
        
        for(int n=2; n<maxn; n+=2)
        {
            // calculate b-e error bound on approximating n/2 voters
            double temp = sqrt(((double)n)/2.0f);
            double bound = BE_C * rho * 2 / (sigma * sigma * sigma * temp);
            
            double pr;

            // if small bound, use approximation
            if(bound < BE_THRESHOLD)
            {
                pr = getApproxProb(randgen,n,alphalist[i],beta[i]);
            }
            else
            {
                //std::cout << "error too high (" << bound << ") at " << n << "\n";
                pr = getExactProb(randgen,n,alphalist[i],beta[i]);
            } 
            pwrongs[i][n/2] = pr;
        }
    }
    
    
    // print it all out
    std::ofstream file;
    file.open(OUTFILE_NAME);
    
    file << "\% " << OUTFILE_NAME << "\n";
    file << "\%\n\% Plotting the chances of getting the wrong 'outcome'\n\% as measured with certain alphas.\n";
    file << "\%\n\n\% Bo Waggoner\n\% Created: ";
    time_t rawtime;
    time(&rawtime);
    file << asctime(localtime(&rawtime)) << "\n%\n\n";
    
    file << "close all;\nclear all;\n\n";
    file << "votes = [" << v[0] << "," << v[1] << "];\n";
    
    file << "mu = " << mu << ";\nsigma = " << sigma << ";\n\n";
    file << "numalpha = " << numalpha << ";\n";
    file << "alphalist = [";
    for(int i=0; i<numalpha; i++)
    {
        file << alphalist[i];
        if(i < numalpha-1) file << ",";
    }
    file << "];\n\n";
    
    file << "beta = [";
    for(int i=0; i<numalpha; i++)
    {
        file << beta[i];
        if(i < numalpha-1) file << ",";
    }
    file << "];\n\n";

    file << "plist = [";
    for(int i=0; i<numalpha; i++)
    {
        for(int n=2; n<maxn; n+=2) {file << pwrongs[i][n/2]; if(n < maxn-2) file << ",";}
        if(i < numalpha-1) file << ";";
    }
    file << "];\n\n";
    
    file << "ns = [2:2:" << (maxn-2) << "];\n\n";
    
    file << "ht = ceil(sqrt(numalpha));\n";
    file << "wd = ceil(numalpha/ht);\n\n";
    
    file << "colorsty = {'k','k','k','k','k','k','k'};\n";
    file << "linesty =  {'-','-','-','-','-','-','-'};\n";
    file << "marksty =  {'*','.','o','+','s','^','x'};\n";
    
    file << "figure(1);\nfor i=1:numalpha,\n    subplot(wd,ht,i);\n";
    file << "    plot(ns,plist(i,:),'LineStyle',linesty{i},'Marker',marksty{i},'Color',colorsty{i});\n";
    file << "    title(['alpha = ',num2str(alphalist(i))]);\n";
    file << "    xlabel('Total number of simulated voters');\n    ylabel('Pr[adjusted m.o.v. \\geq \\beta]');\n";
    file << "    text(" << floor(2*maxn/3) << ",0.55,strcat('beta = ',num2str(beta(i))));\n";
    file << "    axis([0 " << maxn << " 0 0.8]);\n";
    file << "end\n\n";

    file << "skip = 3;\n\n";

    file << "figure(2);\nfor i=1:numalpha,\n";
    file << "    hnew = plot(ns(1:skip:end),plist(i,1:skip:end),'LineStyle',linesty{i},'Marker',marksty{i},'Color',colorsty{i});\n";
    file << "    hold on;\n";
    file << "    legstr = strcat('\\alpha = ',num2str(alphalist(i)));\n";
    file << "    if i==1,\n        legend(legstr);\n    else\n";
    file << "        [LEGH,OBJH,OUTH,OUTM] = legend;\n        legend([OUTH;hnew],OUTM{:},legstr);\n";
    file << "    end\nend\naxis([0 " << maxn << " 0 0.6]);\n\n";
    file << "xlabel('Total number of voters');\n";
    file << "ylabel('Pr[adjusted margin of victory \\geq \\beta]');\n\n";

    for(int i=0; i<numalpha; i++) delete pwrongs[i];
}



